/*  MICROLOGIO RTC (RTC MICROWATCH)

    A watch based on a microcontroller and a LED display with 7-segment numbers
    with COMMON ANODE

    To drive the display without burning the microcontroller's GPIOs, this project
    uses the multiplexing of the output pins.
    
    Hardware used: Atmega168PA or 328P @ 1 MHz with external 32768 Hz crystal
    Fuses set to $62/$D2/$FF (low/high/extended)
    
    Changelog:
    2012/05/25: v. 0.4.2 - Minor fixes
    2012/05/25: v. 0.4.1 - Definitely fix for non-timeout bug
    2012/05/25: v. 0.4.0 - Fixed a bug that left the display steady on even if timeout was over.
                         - Changes at the Prog menu too (now with 2 sub-menus) - inserted the firmware version
    2012/05/14: v. 0.3.3 - Fixed a bug into the power saving routine - now it can run at ~1.5 uA!
    2012/05/13: v. 0.3.2 - A new multiplexing mode
    2012/05/12: v. 0.3.1 - Save&read default settings into internal EEPROM
    2012/05/11: v. 0.3.0 - First release with use of the internal Real-Time Counter
    2012/05/10: v. 0.2.1 - Pover saving mode on/off, set timeout & minor fixes
    2012/05/05: v. 0.2.0 - First stable release
    2012/05/01: v. 0.1.0 - First working release
    2012/04/30: v. 0.0.1 - Very early release
    
    License:
    project released under the terms of the Creative Commons Share Alike-
    -No Commercial-Attribution 3.0 Ita licence
    
    Project by: Leonardo Miliani - www.leonardomiliani.com
*/

//library inclusion
#include <EEPROM.h> //needed to save settings in EEPROM
#include <avr/sleep.h> //needed to use the sleep modes
//the following 2 defines called before PinChange save about 400 byte of Flash memory
#define NO_PORTC_PINCHANGES // to indicate that port c will not be used for pin change interrupts
#define NO_PORTD_PINCHANGES // to indicate that port d will not be used for pin change interrupts
#include <PinChangeInt.h> // needed to use PCINTs


//define the time a segment is ON (it's calculated keeping care of the CPU's clock)
#if (F_CPU==1000000)
#define TIMER 5
#elif (F_CPU==4000000)
#define TIMER 50
#elif (F_CPU==8000000)
#define TIMER 150
#else
#define TIMER 300
#endif

//firmware version
const byte VERSION=0;
const byte REVISION=4;
const byte MINOR_REV=2;

//segments declarations
const byte segA=0;
const byte segB=1;
const byte segC=2;
const byte segD=3;
const byte segE=4;
const byte segF=5;
const byte segG=6;
const byte segH=7;
const byte segments[8] = {segA, segB, segC, segD, segE, segF, segG, segH}; //used to drive the segments
const byte DIG1=A0;
const byte DIG2=A1;
const byte DIG3=A3;
const byte DIG4=A4;
const byte DIG2P=A2;
const byte digits[5] = {DIG1, DIG2, DIG3, DIG4, DIG2P}; // used to drive the digits 

//pins declaration
const byte BUTTON=8; //button to wake up the clock

//this array contains the rapresentation of the numbers in the form of the segments to set to on
//bit values: 0 means ON (because it sets the corresponding pin to LOW so it CAN sink current)
//1 means OFF (because it sets the corresponding pin to HIGH so it CANNOT sink current)
const byte numbers[25] = {
                    0b11000000, // 0:  0
                    0b11111001, // 1:  1
                    0b10100100, // 2:  2
                    0b10110000, // 3:  3
                    0b10011001, // 4:  4
                    0b10010010, // 5:  5
                    0b10000010, // 6:  6
                    0b11111000, // 7:  7
                    0b10000000, // 8:  8
                    0b10011000, // 9:  9
                    0b10001000, // 10: A
                    0b10000011, // 11: B
                    0b11000110, // 12: C
                    0b10100001, // 13: D
                    0b10000110, // 14: E
                    0b10001110, // 15: F
                    0b01111111, // 16: :
                    0b11111111, // 17: empty char
                    0b10001100, // 18: P
                    0b10101111, // 19: r
                    0b10100011, // 20: o
                    0b10010000, // 21: g
                    0b10000111, // 22: t
                    0b10010010, // 23: S
                    0b10100111  // 24: c
};

//global variables
byte clock[5]={0,0,0,0,0}; //this array contains the single numbers to be displayed (the last one is only for colon)
byte clockMode; // display mode
unsigned long timer; //temp variable
byte multiplexing=0; //power saving mode (0=OFF - 1=ON)
byte timeOut=0; //seconds before the display shuts down (default=5)

//temporal variables
volatile unsigned long secs=0;
volatile byte hours=0;
volatile byte minutes=0;
volatile byte secondsX=0;
volatile byte day=0;
volatile byte month=0;
volatile int year=0;
volatile byte daysPerMonth[12] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31};


// *************MAIN CODE ***************

// setup routine
void setup() {
    clockOff(); //set pins direction and initial values
    powerSaver(); //shut down all the internal periphericals not used in this sketch
    pinMode(BUTTON, INPUT); //the input button
    digitalWrite(BUTTON, HIGH); //set internal pull-up to ON
    
    //set initial clock
    loadDefaults();
    //starting time
    secs=0;
    minutes=0;
    hours=0; 
    //starting date
    day=1;
    month=1;
    year=2012; 
    setTimer2(); //prepara il timer 2 per usare il modulo RTC ed il quarzo esterno da 32 kHz    
    

}


//loop
void loop() {
    if (digitalRead(BUTTON)==LOW) { 
        mainLoop();
    }//call the main routine
    PCintPort::attachInterrupt(8, *wakeUp, CHANGE); //attach a PCINT on pin D8
    sleep_mode(); // sleep!
    sleep_disable(); // exit from sleep and set the MCU to full operationality
    PCintPort::detachInterrupt(8); //disable PCINT on button
}


//main routine
void mainLoop() {
    unsigned long tempMillis;
    boolean over;
    
    //initial mode
    clockMode=0; //the clock starts visualizing the time...
    timer=millis()+(timeOut*1000); //...just for a while
    //readTime(); //read the clock
    //clockOn(); //visualize the time
    
    //neverending loop
    do {
        //button pressed? 
        if (digitalRead(BUTTON)==LOW) { 
            delay(50); //little debounce...
            if (digitalRead(BUTTON)==LOW) { //uhm... button still pressed. 
                tempMillis=millis()+1500;
                over=false;
                while (digitalRead(BUTTON)==LOW) {
                    if (millis()>tempMillis) { //the user has pressed the button for more than a specific time
                        over=true;
                        break;
                    }
                }
            }
            if (over == true) { //the user has pressed the button for more than 1 second so he wants to enter into programming mode
                clockMode=1;
                clockOn();
                programClock(); 
            } else { //normal modes
                clockMode++;
                if (clockMode>4) { //if the clock mode is greater than 3 then it shuts down the display
                    clockMode=0;
                }
            }
            if (clockMode>0) { //set a time of xx seconds to display the info
                clockOn();
                timer=millis()+(timeOut*1000);
            }
            delay(150);
        }
    
        
        //which mode?
        switch (clockMode) {
            case 0: //no clock
                //clockOff();
                break;
            case 1: //display time
                stepSeconds(); //alternate ":"
                //display time        
                readTime();
                break;
            case 2: //display date
                clock[4]=17; // no ":"
                readDate();
                break;
            case 3: //display year
                clock[0]=2;
                clock[1]=0;
                clock[4]=17; //no ":"
                getNumbers((byte)(year-2000), 2);
                break;
            case 4: //display seconds
                clock[0]=17;
                clock[1]=17;
                stepSeconds(); //alternate ":"
                getNumbers(secs, 2);
                break;        
        }
    
        //if the timeout is over, shutdown the display
        if (millis()>timer) {
            clockMode=0;
        }
        
        //should I visualize the clock?
        if (clockMode != 0) { 
            displayClock(); 
        }
    } while (clockMode != 0);
    clockOff();    
}


void wakeUp() {
    //empty function. It's only needed because the PCintPort::attachInterrupt function wants it to wake up the MCU
}


//alternate the colons
void stepSeconds() {
    //alternate ":"
    if (secs%2==0) {
        clock[4]=16;
    } else {
        clock[4]=17;
    }
}



//all the segments are set to INPUT (so the clock will go off)
void clockOff() {
    PORTC=0b00000000;
    DDRC=0b00000000;
    PORTD=0b00000000;
    DDRD=0b00000000;
}


//initialize the pins
void clockOn() {
    PORTC=0b00000000; //LOW on A0..A7 - so no output current to display
    DDRC=0b00011111; //set pins A0..A4 as outputs
    PORTD=0b11111111; //HIGH on D0..D7 - pins are not shortcircuited because LEDs are like diodes    
    DDRD=0b11111111; //set pins D0..D7 as output
}


//read day&month from the software RTC
void readDate() {
    byte temp1=day;
    byte temp2=month;
    
    //divide into 2 digits
    getNumbers(temp1, 0);
    getNumbers(temp2, 2);
}


//read hours&minutes from the software RTC
void readTime() {
    byte temp1=hours;
    byte temp2=minutes;
    
    //divide into 2 digits
    getNumbers(temp1, 0);
    getNumbers(temp2, 2);
}


//put the number inside a couple of digits
void getNumbers(byte number, byte index) {
    byte temp;
    
    //add a "0" if the number is lesser than 10
    if (number>9) {
        temp=int(number/10);
        clock[index]=temp;
        number-=(temp*10);
    } else {
        clock[index]=0;
    }
    clock[index+1]=number;
}


//show on the display the content of the array (clock, date, seconds or a message)
void displayClock() {
    byte temp;
    byte max;

    //multiplexing: 3 schemes
    if (multiplexing == 2) { //each time, a single segment per single digit is lighted on
        for (byte i=0; i<8; i++) {
            for (byte j=0; j<5; j++) {
                temp=numbers[clock[j]];
                digitalWrite(digits[j], HIGH);
                digitalWrite(segments[i], bitRead(temp, i)); //first is turned on...
                delayMicroseconds(TIMER); //.. then a little time is waited....
                digitalWrite(segments[i], ~bitRead(temp, i)); //...then the segment is turned off.
                digitalWrite(digits[j], LOW);
            }
        }
    } else {
        for (byte j=0; j<5; j++) { //a digit per time
            digitalWrite(digits[j], HIGH);
            temp=numbers[clock[j]];
            if (multiplexing == 1) { //multiplexing of single segments, from the first digit to the last one
                for (byte i=0; i<8; i++) { //a segment per time
                    digitalWrite(segments[i], bitRead(temp, i)); //first is turned on...
                    delayMicroseconds(TIMER); //.. then a little time is waited....
                    digitalWrite(segments[i], ~bitRead(temp, i)); //...then the segment is turned off.
                }
            } else { //multiplexing of entire digits at a time
                PORTD = temp;
                DDRD = 0b11111111;
                delayMicroseconds(TIMER*2);
                DDRD = 0b00000000;
            }            
            digitalWrite(digits[j], LOW); //turn off the entire digit
        }
    }
}


//the user can set the RTC
void programClock() {
    unsigned long tempMillis;
    byte tempClock[5] = {0, 0, 0, 0, 12};
    
    //show "Prog" on display
    clock[0]=18; //P
    clock[1]=19; //r
    clock[2]=20; //o
    clock[3]=21; //g
    clock[4]=17;
    
    tempMillis=millis()+750;
    while (tempMillis>millis()) {
        displayClock();
    }    
    
    byte temp=chooseMode();
    switch (temp) {
        case 0:
            //the user wants to change the time&date
            //be careful: NO check is done to control if the datas are valids!
    
            //set hours
            delay(750);
            clock[0]=0;
            clock[1]=0;
            clock[2]=17;
            clock[3]=17;
            clock[4]=16;
            tempClock[0]=readNumber(23, 0, 0);
        
            //set minutes
            delay(750);    
            clock[0]=17;
            clock[1]=17;
            clock[2]=0;
            clock[3]=0;
            tempClock[1]=readNumber(59, 0, 2);
        
            //continue programing the clock with DATE
            //set day
            delay(750);
            clock[0]=0;
            clock[1]=0;
            clock[2]=17;
            clock[3]=17;
            clock[4]=17;
            tempClock[2]=readNumber(31, 1, 0);
        
            //set month
            delay(750);
            clock[0]=17;
            clock[1]=17;
            clock[2]=0;
            clock[3]=0;
            tempClock[3]=readNumber(12, 1, 2);
    
            //set year
            delay(750);
            clock[0]=2;
            clock[1]=0;
            clock[2]=1;
            clock[3]=0;
            tempClock[4]=readNumber(29, 10, 2);    
    
            //set the software RTC
            delay(750);
            //halts the RTC
            SREG &= ~(1<<SREG_I);
            hours=tempClock[0];
            minutes=tempClock[1];
            secs=0;
            day=tempClock[2];
            month=tempClock[3];
            year=2000+tempClock[4];
            TCNT2=0; //reset the timer to 0 
            TIMSK2 |= (1<<TOIE2); //set an interrupt signal on timer's overflow
            SREG|=1<<SREG_I; //start monitoring the interrupts
            break;
        case 1:
            //the user wants to program the preferences
            //set power saving mode
            delay(750);
            clock[0]=23; // "S"
            clock[1]=24; // "c"
            clock[2]=0;
            clock[3]=0;
            multiplexing=readNumber(2, 0, 2);
            
            //set the timeout
            delay(750);
            clock[0]=22; // "t"
            clock[1]=20; // "o"
            clock[2]=0;
            clock[3]=1;
            timeOut=readNumber(15, 1, 2);
        
            //save the preferences into EEPROM
            saveDefaults();
            break;
        case 2:
            //let's display the firmware version
            clock[0]=17;
            clock[1]=VERSION;
            clock[2]=REVISION;
            clock[3]=MINOR_REV;
            tempMillis=millis()+3000;
            while (tempMillis>millis()) {
                displayClock();
            }   
            break;
    }
}


//set the clock's setup
byte chooseMode() {
    unsigned long tempMillis;
    boolean flag=false;
    byte counter=0;
    int timeInterval=1000;
    
    //timeout before to increment the value of the counter
    tempMillis=millis()+timeInterval;
    clock[4]=17;
    do {
        switch (counter) {
            case 0:
                clock[0]=13; //"D"
                clock[1]=10; //"A"
                clock[2]=22; //"t"
                clock[3]=14; //"E"
                break;
            case 1:
                clock[0]=23; //"S"
                clock[1]=14; //"E"
                clock[2]=22; //"t"
                clock[3]=17;  //""
                break;
            case 2:
                clock[0]=23; //"S"
                clock[1]=20; //"o"
                clock[2]=15; //"F"
                clock[3]=22; //"t"
                break;
        }
        displayClock();
        //has the user pressed the button?
        if (digitalRead(BUTTON)==LOW) {
            delay(100);
            if (digitalRead(BUTTON)==LOW) {
                //yes: let's exit
                flag=true;
                break;
            }
        }
        //timeout is over: we have to increment the digit to be displayed
        if (millis()>tempMillis) {
            counter++;
            tempMillis=millis()+timeInterval;
            if (counter>2) {
                counter=0;
            }
        }
    } while (!flag);
    return counter; //return the user's choice
        

}


//read a number from the input interface
byte readNumber(byte limit, byte startValue, byte cell) {
    unsigned long tempMillis;
    boolean flag=false;
    byte counter=startValue;
    int timeInterval=650;
    
    //xxx ms before to increment the value of the counter
    tempMillis=millis()+timeInterval;
    do {
        getNumbers(counter, cell);
        displayClock();
        //has the user pressed the button?
        if (digitalRead(BUTTON)==LOW) {
            delay(100);
            if (digitalRead(BUTTON)==LOW) {
                //yes: let's exit
                flag=true;
                break;
            }
        }
        //timeout is over: we have to increment the digit to be displayed
        if (millis()>tempMillis) {
            counter++;
            tempMillis=millis()+timeInterval;
            if (counter>limit) {
                counter=startValue;
            }
        }
    } while (!flag);
    return counter; //return the user's choice
}


//maximum power saving! disable all the unused periphericals
void powerSaver() {
    //set the internal registers to reduce power consumes
    PRR &= ~((1<<PRUSART0) | (1<<PRADC) | (1<<PRSPI) | (1<<PRTIM1) | (PRTWI)); //shut down the following:  USART, ADC, I2C, Timer1, SPI
    ADCSRA &= ~(1<<ADEN); //shut down the ADC
    ACSR = (1<<ACD); //shut down the analog comparator
    DIDR0 = ((1<<ADC5D) | (1<<ADC4D) | (1<<ADC3D) | (1<<ADC2D) | (1<<ADC1D) | (1<<ADC0D)); //shut down digital input buffers on pins ADC0..5
    DIDR1 = (1<<AIN1D)|(1<<AIN0D); //shut down digital input buffer on AIN0..1
    MCUCR |= (1<<BODS); //BOD disabled
    WDTCSR &= ~((1<<WDCE) | (1<<WDE)); //watchdog disabled

    //prepare the sleep mode
    set_sleep_mode(SLEEP_MODE_PWR_SAVE); // set the sleep for POWER SAVE mode
    sleep_enable(); // sleep ready to use
}


//set timer 2 to use it in combination with the Real-Time Counter module
void setTimer2() {
    //all the interrupts on timer 2 to OFF
    TIMSK2 &= ~(1<<TOIE2); 
    TIMSK2 &= ~((1<<OCIE2A) | (1<<OCIE2B));
    //set the use of the external crystal as source clock
    ASSR |= (1<<AS2);
    //Prescaler/128: thanks to this, with a 32768 Hz crystal we get an overflow after 1 second
    TCCR2B |= ((1<<CS22) | (1<<CS20)); 
    TCCR2B &= ~(1<<CS21); 
    //Counter mode: the timer is incremented until overflow
    TCCR2A &= ~((1<<WGM21) | (1<<WGM20)); 
    TCCR2B &= ~(1<<WGM22);

    //set the initial value of the timer
    TCNT2=0;
    //set an interrupt signal at the timer's overflow
    TIMSK2 |= (1<<TOIE2); 
    //start monitoring the interrupt
    SREG|=1<<SREG_I;
}


//Interrupt Routine: this manipulates the software RTC
ISR(TIMER2_OVF_vect) {

    byte dayT;
    secs++; //increment seconds
    if (secs>59) { //1 minute has passed
        secs=0;
	minutes++;
	if (minutes>59) { //1 hour has passed
	    minutes=0;
	    hours++;
	    if (hours>23) { //1 day has passed
	        hours=0;
		day++;
		if (month == 2) { //february?
		    if (isLeapYear(year)) { //leap year?
		        dayT=29;
		    } else {
			dayT=28;
		    }
		} else {
		    dayT=daysPerMonth[month-1];
		}
		if (day>dayT) { //is this the last day of the month?
		    day=1;
		    month++;
		    if (month>12) { //yeah: 1 year more
		        month=1;
			year++;
		    }
		}
	    }
	}
    }
}


//function to check leap years
boolean isLeapYear(int yearT) {
    if (((yearT % 4) == 0) && ((yearT % 100) != 0) || ((yearT % 400) == 0)) {
        return true;
    } else {
        return false;
    }
}


//load default settings from interal EEPROM
void loadDefaults() {
    byte tempByte;
    
    tempByte=EEPROM.read(0);
    if (tempByte != 42) { //first time
        EEPROM.write(0, 42);
        EEPROM.write(1, 1); // power saving ON
        multiplexing=1;
        EEPROM.write(2, 5); // timeout = 5
        timeOut=5;
    } else {
        multiplexing=EEPROM.read(1);
        timeOut=EEPROM.read(2);
    }
}


//save default settings into the internal EEPROM
void saveDefaults() {
    EEPROM.write(0, 42); //to be sure that the program will recognize the values the next restart
    EEPROM.write(1, multiplexing); // power saving
    EEPROM.write(2, timeOut); // timeout
}
